#!/bin/bash
#  This  simple script helps deleting hosts from the user's known_hosts
#  file when their pubkey has changed.
#    Copyright (C) 2006-2009 by Phil Sutter <phil@nwl.cc>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


### VARIABLE DEFINITION PART

# static variables, changing them should not be neccessary
ME="$(basename $0)"
SSH_CNF=$HOME/.ssh/config
KH_FILE=$HOME/.ssh/known_hosts

# variable holding the real host found by search_in_sshconf()
REAL_HOST=""

# variables holding the host information filled in by lookup_hostip()
LU_HOST=""
LU_IP=""

# work counter
WRKCOUNT=0

### FUNCTION DEFINITION PART

usage() { # no args
	# print usage information
	echo -n "Usage: $ME [${ME#un}-opts (ignored)]"
	echo " [user@]host[:[path]] [${ME#un}-opts (ignored)]"
}
printstat() { # (return_code)
	# wordify return codes
	if [[ $1 -eq 0 ]]; then
		echo "OK"
	else
		echo "FAIL"
	fi
	return $1
}
search_in_sshcnf() { # (expr)
	grep -qi "^host $1$" "$SSH_CNF" || return 1
	# the code below  calls awk in three steps:
	# * filter empty and commented out lines
	# * divide content into records containing a single "Host" definition
	# * search for the "Hostname" in the resulting strings and print the value
	REAL_HOST="$(grep -v -e '^$' -e '^[[:space:]]*#.*$' "$SSH_CNF" | \
			awk -v IGNORECASE=1 -v RS="host " "/^$1[[:space:]]*/" | \
			awk -v IGNORECASE=1 '/hostname/{print $2}')"

	[[ -n $REAL_HOST ]] || return 1
	echo "found realname $REAL_HOST for alias $1"
	return 0
}

std_search_in_khfile() { # (expr)
	# search for the given expr in the khfile
	local count
	[[ -n $1 ]] || return 1
	count=$(grep -c -i "$1 " "$KH_FILE")
	if [[ $count -gt 1 ]]; then
		echo -n "Something strange is going on here,"
		echo " you seem to have more than one entry for"
		echo "this host ($1) in your known_hosts file. "
	fi
	return $([ $count -eq 1 ])
}
kgen_search_in_khfile() { # (expr)
	# search for the given expr in the khfile
	[[ -n $1 ]] || return 1
	if [[ -z "$(ssh-keygen -F "$1")" ]]; then
		return 1
	else
		return 0
	fi
}
std_seek_and_destroy() { # (expr)
	# delete lines of the khfile containing expr
	cp "$KH_FILE" "${KH_FILE}.old"
	printf "/%s /\nd\nwq\n" "$1" | ed -s "$KH_FILE" >/dev/null 2>&1
	return $?
}
kgen_seek_and_destroy() { # (expr)
	# use ssh-keygen to remove matching hostnames
	ssh-keygen -R "$1" >/dev/null 2>&1
	return $?
}
lookup_hostip() { # (name) to lookup. sets $LU_HOST, $LU_IP
	local name
	# check first if given name is an alias
	if search_in_sshcnf "$1"; then
		name="$REAL_HOST"
	else
		name="$1"
	fi
	# does a host lookup to get the corresponding IP/hostname
	eval "$(hostx "$name" 2>&1 | \
	        awk '/exist/{exit 1} \
		     /Name:/{print "LU_HOST="$2} \
		     /Address:/{print "LU_IP="$2} \
		     /^'"$name"'/{printf "LU_HOST=%s\nLU_IP=%s",$1,$3}' \
	       )"
	if [ -z "$LU_HOST" ]; then
		if [[ "$name" = *.*.*.* ]]; then
			LU_IP="$name"
		else
			LU_HOST="$name"
		fi
	fi
}

### PROGRAM ENTRY POINT

# check arguments count
[[ $# -lt 1 ]] && { usage ; exit 1; }

# Do a fast lookup in case there is an '@' or ':' identifying the
# argument containing the hostname/IP to search for. Note that this
# algorithm is also correct if both '@' and ':' occur.
found=false
ignore=false
for arg in $@; do
	if $ignore; then
		ignore=false
		continue
	fi
	case $arg in
	(-D|-L|-R|-w)
		# skip args which may contain ':' or '@'
		ignore=true
		continue
		;;
	(*@*)
		arg=${arg#*@}
		found=true
		;;
	(*:*)
		arg=${arg%:*}
		found=true
		;;
	esac
	if $found; then
		lookup_hostip $arg
		break
	fi
done

# unscp: if above failed, syntax was wrong!
if [[ $ME == unscp ]]; then
	if ! $found; then
		echo "check your syntax, could not find any param containing a hostname!"
		usage
		exit 1
	fi
fi

# ignore all non-hostname args that are valid for ssh,
# so users can use their shell history to run this command
if [[ -z $LU_HOST && -z $LU_IP ]]; then
	while true; do
		case "$1" in
		-1|-2|-4|-6|-A|-a|-C|-f|-g|\
		-K|-k|-M|-N|-n|-q|-s|-T|-t|-V|\
		-v|-X|-x|-Y|-y)
			shift
		;;
		-b|-c|-D|-e|-F|-i|-L|-l|-m|\
		-O|-o|-R|-S|-w)
			shift 2
		;;
		-p)
			shift
			[[ $ME == unssh ]] && shift
		;;
		*)
			lookup_hostip "$1"
			break
		;;
		esac
	done
fi

# check whether ssh-keygen(1) is available, and set function prefix
# accordingly
if which ssh-keygen >/dev/null 2>&1; then
	PFX="kgen"
else
	PFX="std"
fi

# manually backup here, so multiple changes won't harm the history
cp "$KH_FILE" "${KH_FILE}.unssh_bak"

# search for the hostname if existent
if ${PFX}_search_in_khfile "$LU_HOST"; then
	echo -n "Found $LU_HOST, deleting..."
	${PFX}_seek_and_destroy "$LU_HOST"
	printstat $?
fi

# search also for a short version of the hostname
if [[ $LU_HOST = *.* ]]; then
	LU_HOST=$(awk -v FS=. '{print $1}' <<< $LU_HOST)
	if ${PFX}_search_in_khfile "$LU_HOST"; then
		echo -n "Found $LU_HOST, deleting..."
		${PFX}_seek_and_destroy "$LU_HOST"
		printstat $?
	fi
fi

# search for the IP which might not be found in case the
# searches above already deleted the corresponding line
if [[ -n $LU_IP ]] && ${PFX}_search_in_khfile "$LU_IP"; then
	echo -n "Found $LU_IP, deleting..."
	${PFX}_seek_and_destroy "$LU_IP"
	printstat $?
fi

# use the backup to count the number of entries removed
WRKCOUNT=$(expr $(wc -l < "${KH_FILE}.unssh_bak") - $(wc -l < "$KH_FILE"))

if [[ $WRKCOUNT -gt 0 ]]; then
	echo "all done, deleted $WRKCOUNT entries"
	mv "${KH_FILE}.unssh_bak" "${KH_FILE}.old"
	exit 0
else
	echo "nothing found, so nothing done"
	rm "${KH_FILE}.unssh_bak"
	exit 1
fi

### this is the end - my only friend, the end
